#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <string.h>      //strlen()
#include <strings.h>     //bzero()
#include <pthread.h>
#include <signal.h>             //for sigaction(),
#include <poll.h>
#include <unistd.h>

#define BUFFER 4096
#define CLIENT_PORT 5000
#define MAX_SERVERS 100
#define SERVER_PORT 80
#define MAX_QUEUE_SIZE 10
#define SLEEP_TIME 10

typedef struct serverLoad ServerStats;
struct serverLoad
{
	int serverid;	//unique Id given by this program automatically 0,1,2..
	char ip[BUFFER];//IP address of backend server
	int status;	//backend server is up or down
	double min[3];	//min[0]=loadAvg 1 minute before, min[1]= before 5 minute & min[3] = before 15 minute
};
typedef struct connection Connection;
struct connection
{
	int client_fd;	//file_descriptor of client
	int server_fd;	//file_descriptor of server
};
typedef struct serverArgs ServerArgs;
struct serverArgs	//Used to pass arguments to thread function
{
	ServerStats *s;
	char *filename;
	int max_servers;
};

char * getLoadAvg(char *ip, int port);
int getServerConnection(char *ip, int port);
int updateServerStats(ServerStats s[], int max_servers, char *filename);
static int cmpLoad(const void *a, const void *b);
void *connection1_to_connection2_copier (void *arg);
void *connection2_to_connection1_copier (void *arg);
void *serverStatsUpdater(void *arg);

/** It is main function. It updates backend servers' load average statistics at 
  * regular interval using thread. It accepts http request from users, and serves 
  * the request with a backend server which has minimum load average.
  * @param config_file_name as command line argument, file contains list of backend servers
  *
  */
int main(int argc, char * argv[])
{
	ServerStats s[MAX_SERVERS];
	int serverCount;
	//char * loadStr;
	int i;
	//Thread + socket 
	struct sockaddr_in server_address;
	int return_value;
	//struct sigaction act1;
    	pthread_t child1;
	pthread_t child2;
	pthread_t serverUpdater;
    	void *child_return_value;
	Connection con;	//to be passed in thread
	ServerArgs server_args;
	int listen_file_descriptor;
	int connection_file_descriptor;

	if(argc<2)
	{
		printf("usage:argv[0] <config_file>\n");
		return 0;
	}
	//updating server statistics from file
	serverCount = updateServerStats(s, MAX_SERVERS, argv[1]);
	printf("ServerCount:%d, filename:%s\n", serverCount, argv[1]);
	//Creating a thread which will update server status at regular interval
	server_args.s = s;
	server_args.filename = argv[1];
	server_args.max_servers = MAX_SERVERS; 
        return_value =  pthread_create(&serverUpdater, NULL, serverStatsUpdater, &server_args);
        if(return_value !=0)
        {
            perror("Thread  creation failed for ServerUpdater\n");
            exit(EXIT_FAILURE);
        }
	
	//listening on LISTEN_PORT1
    listen_file_descriptor = socket (AF_INET, SOCK_STREAM, 0);
    if (listen_file_descriptor < 0)
    {
        fprintf (stderr, "%s: cannot open socket.\n", argv[0]);
        exit (EXIT_FAILURE);
    }

    bzero (&server_address, sizeof (server_address));
    server_address.sin_family = AF_INET;
    server_address.sin_addr.s_addr = htonl (INADDR_ANY);
    server_address.sin_port = htons (SERVER_PORT);
    return_value =
        bind (listen_file_descriptor, (struct sockaddr *) &server_address,
              sizeof (server_address));
    if (return_value < 0)
    {
        perror ("Cannot bind");
        exit (EXIT_FAILURE);
    }

    return_value = listen (listen_file_descriptor, MAX_QUEUE_SIZE);
    if (return_value < 0)
    {
        perror ("Cannot listen");
        exit (EXIT_FAILURE);
    }

    while (1)
    {
        //Get first connection
        connection_file_descriptor =
            accept (listen_file_descriptor, (struct sockaddr *) NULL, NULL);
        if (connection_file_descriptor < 0)
        {
            perror ("accept() failed.");
            exit (EXIT_FAILURE);
        }
        else
        {
            printf ("Got client for connection.\n");
        }
	//Create one threads to pipe the two connections
	if(s[0].status==0)
	{
		printf("No server is available..\n");
		continue;
	}
	con.client_fd = connection_file_descriptor;
	for(i=0; i<serverCount; i++)
	{
		con.server_fd = getServerConnection(s[0].ip, SERVER_PORT);
		if(con.server_fd > 0)	//valid connection
			break;
	}
        return_value =  pthread_create(&child1, NULL, connection1_to_connection2_copier, &con);
        if(return_value !=0)
        {
            perror("Thread 1 creation failed");
            exit(EXIT_FAILURE);
        }

	return_value =  pthread_create(&child2, NULL, connection2_to_connection1_copier, &con);
        if(return_value !=0)
        {
            perror("Thread 2 creation failed");
            exit(EXIT_FAILURE);
        }

	//Wait for childs to finish
        return_value=pthread_join(child1, &child_return_value);
        if(return_value !=0)
        {
            perror("Thread 1 join failed");
            exit(EXIT_FAILURE);
        }
        return_value=pthread_join(child2, &child_return_value);
        if(return_value !=0)
        {
            perror("Thread 2 join failed");
            exit(EXIT_FAILURE);
        }

	//TODO: Move it to thread function
	//close connections, should only be done once, hence preferably in parent
        return_value = close (con.server_fd);
        if (return_value < 0)
        {
            perror ("close failed");
            exit (EXIT_FAILURE);
        }
        return_value = close (con.client_fd);
        if (return_value < 0)
        {
            perror ("close failed");
            exit (EXIT_FAILURE);
        }

    }
	return 0;
}

/** It will update load avg. statistics of each server.
  * @param s[] Array of ServerStats structure to store server statistics
  * @param max_servers size of array or max possible servers
  * @param filename Name/path of file where list of IP of backend servers is available
  *
  */
int updateServerStats(ServerStats s[], int max_servers, char *filename)
{
	FILE *fp;
	int count = 0, i=0, j;
	char *loadLine;
	char * strp;
	fp = fopen(filename,"r");
	if(fp == NULL) {printf("Unable to open file %s\n", filename);return 0;}
	while(!feof(fp))
	{
		s[i].serverid = i;
		fscanf(fp, "%s", s[i].ip);
		if(strlen(s[i].ip)<7 || strlen(s[i].ip)>15)
			continue;
		loadLine = getLoadAvg(s[i].ip, CLIENT_PORT);
		strp = loadLine;
		if(loadLine==NULL)
			s[i].status = 0;
		else
			s[i].status = 1;
		j = 0;
		while(strp && *strp)	//strp!=NULL && strp[0]!='\0'
		{
			strp++;
			s[i].min[j++] = atof(strp);
			while(strp && *strp!=',' && *strp!='\0')
				strp++;
		}
		printf("id:%d,ip:%s,status:%d,load:%.2f,%.2f,%.2f\n", s[i].serverid, s[i].ip, s[i].status, s[i].min[0], s[i].min[1], s[i].min[2]);
		i++;	//next server
	}
	count = i;
	qsort(s, count, sizeof(ServerStats), cmpLoad);
	for(i=0; i<count; i++)
	{
		printf("%d ",s[i].serverid);
	}
	printf("\n");
	return count;
}

/** It connects to load_client on backend servers, and gets load avg statistics
  * @param ip IP address of backend server
  * @param port Port of load_client where it is listening.
  *
  */
char * getLoadAvg(char *ip, int port)
{
	static char loadAvg[BUFFER];
	int bytes_read;
	int host_file_descriptor;
	host_file_descriptor = getServerConnection(ip, port);
	if(host_file_descriptor < 0) return NULL;	
	bytes_read = read(host_file_descriptor, loadAvg, BUFFER);
	if(bytes_read < 0) 
	{
		fprintf(stderr, ": cannot read from socket.\n");
		return NULL;
	}
	return loadAvg;
}

/** It will return file_descriptor of a socket, if ip and port of destination
  * is provided.
  * @param ip
  * @param port
  *
  */
int getServerConnection(char *ip, int port)
{
	int host_file_descriptor;
	struct sockaddr_in host_addr;
    	struct hostent *server;
	//procedure of coonecting to host
        //try to open socket
        host_file_descriptor = socket(AF_INET, SOCK_STREAM, 0);

        //if socket opening failed then print error message and quit
        if(host_file_descriptor < 0)
        {
                fprintf(stderr, ": cannot open socket.\n");
                return -1;
        }
	
	server = gethostbyname(ip);
	if(server == NULL)
	{
                fprintf(stderr, ": the supplied ipv4 address  is incorrect.\n");
                return -1;
        }

        //initialize struct server_address to all zeros.
        bzero(&host_addr, sizeof(host_addr));

        //set the protocol used to be ipv4
        host_addr.sin_family = AF_INET;

	bcopy((char *)server->h_addr, 
         (char *)&host_addr.sin_addr.s_addr,
         server->h_length);

        //set the port to be 80 0r 8080 stored in host_port
        host_addr.sin_port = htons(port);

        //connecting to host
        if(connect(host_file_descriptor, (struct sockaddr *) &host_addr, sizeof(host_addr)) < 0)
        {
                fprintf(stderr, ": cannot connect to host %s\n", ip);
                return -1;
        }

	return host_file_descriptor;
}

/** It is comparision function for qsort(). 
  * It is decesion making function of the load_balancer.
  * It gives preferance to server, which is up.
  * then, to compare load average of two servers, I have used
  * 15*load1 + 5*load5 + 1*load15 should be minimum, where load15 is load before 15 minute.
  *
  */
static int cmpLoad(const void *a, const void *b)
{
	ServerStats *p,*q;
	p = (ServerStats *)a;
	q = (ServerStats *)b;
	double load1, load2;
	if(p->status==0) return 1;
	if(q->status==0) return -1;
	load1 = 15*p->min[0] + 5*p->min[1] + p->min[2];
	load2 = 15*q->min[0] + 5*q->min[1] + q->min[2]; 
	return load1-load2;
}

/** It is thread function, which copies data from client to server.
  *
  */
void *connection1_to_connection2_copier (void *arg)
{
    Connection *con = (Connection *)arg;
    int client_fd = con->client_fd;
    int server_fd = con->server_fd;
    struct pollfd pollfd1[1];
    char input_data[BUFFER];
    int characters_read;
    int characters_written;

    pollfd1[0].fd = server_fd;
    pollfd1[0].events = POLLOUT;

    while (1)
    {
        characters_read = read (client_fd, input_data, BUFFER - 1);
        if (characters_read > 0)
        {
            input_data[characters_read] = '\0';
            //printf ("Received %d bytes from client : %s\n", characters_read, input_data);
            characters_written = write (server_fd, input_data, characters_read);
            poll (pollfd1, 1, 0);
            if (pollfd1[0].revents & POLLHUP)
            {
                printf ("Connection2 seems to be closed. Ending this thread..\n");
                break;
            }
            printf ("Wrote %d bytes to connection2\n", characters_written);
        }
        else if (characters_read < 0)
        {
            perror ("Read from connection1 failed");
            exit (EXIT_FAILURE);
        }
        else
        {
            // characters_read=0, and return on blocking IO means connection1 closed.
            printf ("Connection1 seems to be closed. Ending this thread.\n");
            break;
        }
    }

    pthread_exit(NULL);
    return NULL;
}

/** It is thread function, which copies data from server to client.
  *
  */
void *connection2_to_connection1_copier (void *arg)
{
    Connection *con = (Connection *)arg;
    int client_fd = con->client_fd;
    int server_fd = con->server_fd;
    struct pollfd pollfd1[1];
    char input_data[BUFFER];
    int characters_read;
    int characters_written;

    pollfd1[0].fd = client_fd;
    pollfd1[0].events = POLLOUT;

    while (1)
    {
        characters_read = read (server_fd, input_data, BUFFER - 1);
        if (characters_read > 0)
        {
            input_data[characters_read] = '\0';
            //printf ("Received %d bytes from client : %s\n", characters_read, input_data);
            characters_written = write (client_fd, input_data, characters_read);
            poll (pollfd1, 1, 0);
            if (pollfd1[0].revents & POLLHUP)
            {
                printf ("Connection1 seems to be closed. Ending this thread..\n");
                break;
            }
            printf ("Wrote %d bytes to connection1\n", characters_written);
        }
        else if (characters_read < 0)
        {
            perror ("Read from connection2 failed");
            exit (EXIT_FAILURE);
        }
        else
        {
            // characters_read=0, and return on blocking IO means connection1 closed.
            printf ("Connection2 seems to be closed. Ending this thread.\n");
            break;
        }
    }

    pthread_exit(NULL);
    return NULL;
}

/** It is thread function, which updates server statistics at regular interval
  * defined by SLEEP_TIME.
  *
  */
void *serverStatsUpdater(void *arg)
{
	ServerArgs *server_args = (ServerArgs *)arg;
	ServerStats *s = server_args->s;
	char *filename = server_args->filename;
	int max_servers = server_args->max_servers;
	int count;
	while(1)
	{
		count = updateServerStats(s, max_servers, filename);
		sleep(SLEEP_TIME);
	} 
	pthread_exit(NULL);
	return NULL;
}

